#include <config.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#include "args.h"
#include "bibyydcl.h"

#define	ERROR_PREFIX	"??"	/* this prefixes all error messages */

extern long		line_number;
extern const char	*the_filename;

static void		dolex ARGS((void));
int			main ARGS((int argc_, char *argv_[]));
static int		match_option ARGS((const char *s, const char *t, int nmax));
static void		show_help_and_exit ARGS((void));
static void		show_version_and_exit ARGS((void));

static const char * PROGRAM_NAME = "biblex";

#undef MAX
#define MAX(a,b)	(((a) >= (b)) ? (a) : (b))

#undef TOLOWER
#define TOLOWER(c)	(isupper(c) ? tolower(c) : (c)) /* guard against broken tolower() on ancient systems */


#if NEW_STYLE
int
main(int argc, char *argv[])
#else /* K&R style */
int
main(argc, argv)
int argc;
char *argv[];
#endif /* NEW_STYLE */
{
    int k;
    FILE *fp;

    PROGRAM_NAME = argv[0];

    if (argc > 1)		/* files to lex named on command line */
    {
	for (k = 1; k < argc; ++k)
	{
	    const char *this_arg;

	    this_arg = argv[k];

	    /* Accept both "-arg" and GNU-style "--arg" */
	    if ( (this_arg[0] == '-') && (this_arg[1] == '-') )
		this_arg++;			/* change "--arg" to "-arg" */

	    if (match_option(this_arg, "-help", MAX(2, strlen(this_arg))))
		show_help_and_exit();
	    else if (match_option(this_arg, "-version", MAX(2, strlen(this_arg))))
		show_version_and_exit();
	    else
	    {
		fp = freopen(argv[k],"r",stdin);
		if (fp == (FILE*)NULL)
		{
		    (void)fprintf(stderr,
			"\n%s Ignoring open failure on file [%s]\n",
			ERROR_PREFIX, argv[k]);
		    (void)perror("perror() says");
		}
		else
		{
		    the_filename = (const char*)argv[k];
		    dolex();
		    (void)fclose(fp);
		}
	    }
	}
    }
    else				/* lex stdin */
    {
	the_filename = "stdin";
	dolex();
    }
    exit (EXIT_SUCCESS);
    return (0);
}


static void
dolex(VOID)
{
    line_number = 1L;
    printf("# line %ld \"%s\"\n", line_number, the_filename);

    while (yylex() > 0)
	;
}


#if NEW_STYLE
static int
match_option(const char *s, const char *t, int nmax)
#else /* K&R style */
static int
match_option(s, t, nmax)
const char *s;
const char *t;
int nmax;
#endif /* NEW_STYLE */
{
    int match, n;

    match = 1;		/* assume that s and t match, and then try to prove otherwise */

    for (n = 0; n < nmax; ++n)
    {
	if (TOLOWER(s[n]) != TOLOWER(t[n]))
	{
	    match = 0;
	    break;
	}
    }

    return (match);
}


static void
show_help_and_exit(VOID)
{
    static const char *help[] =
    {
	"\n",
	"Usage:\n\t",
	"REPLACE-ME",
	" [ options ] [ infile(s) ]\n",
	"\n",
	"If there are no command-line files, then input is from stdin.\n",
	"\n",
	"Optional arguments:\n",
	"\n",
	"\t-h\tprint brief help on stderr and exit with success\n",
	"\t-v\tprint version information on stderr and exit with success\n",
	"\n",
	"The GNU-style --help, and --version options are accepted, as are\n",
	"any unique prefixes of those long options.  Lettercase is not\n",
	"significant in option names.\n",
	"\n",
	(const char *)NULL
    };
    int k;

    help[2] = PROGRAM_NAME;

    for (k = 0; help[k] != (const char *)NULL; ++k)
	(void)fprintf(stderr, "%s", help[k]);

    exit(EXIT_SUCCESS);
}


static void
show_version_and_exit(VOID)
{
    const char *version[] =
    {
	"This is ",
	"REPLACE-ME",
	" version ",
	PACKAGE_VERSION,
	" ",
	PACKAGE_DATE,
	"\n\n",
	"Copyright (C) 1998--2014 by ",
	PACKAGE_AUTHOR,
	" and the Free Software Foundation, Inc.\n\n",
	PACKAGE_NAME,
	" is free software that you are free to change, redistribute, and\n",
	"mirror.  There is NO WARRANTY; not even for MERCHANTABILITY or FITNESS\n",
	"FOR A PARTICULAR PURPOSE.\n\n",
	"The software development master Web site is\n\n\t",
	PACKAGE_URL,
	"\n\n",
	"The software is distributed under the GNU General Public License (GPL),\n",
	"version 2 or later, available as the file COPYING in the program source\n",
	"distribution, and also at\n\n",
	"\thttp://www.gnu.org/copyleft/gpl.html\n\n",
	"Send bug reports to\n\n\t",
	PACKAGE_BUGREPORT,
	"\n\n",
	(const char *)NULL
    };
    int k;

    version[1] = PROGRAM_NAME;
    for (k = 0; version[k] != (const char *)NULL; ++k)
	(void)fprintf(stderr, "%s", version[k]);

    exit(EXIT_SUCCESS);
}
